/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 571 $ $LastChangedDate: 2002-08-25 22:24:49 +0200 (Sun, 25 Aug 2002) $
 *
 * Type - Type checker.
 */

/*
TODO: It fails on following code:
void ble()
{
	_ pow(_ mul, _ one)
	{
		_ f(_ val, _ n) {
			if (n == 0)
				return one;
				
			if (n == 1)
				return val;
			
			_ res = f(val, n / 2);

			res = mul(res, res);
			
			if (n % 2 == 1)
				res = mul(res, val);

			return res;
		}

		return f;
	}
}
*/

open Ast;
open Util;
open Type;

exception string Type_error;

string string_of_type(ty t)
{
    string string_of_field(<<'a>symbol>field f)
    {
	return string_of_type(f.fld_type) + " " + qualified_name(f.fld_symbol);
    }
    
    string record_content(<<<'a>symbol>field>List::t lst)
    {
	return String::concat("; ", List::map(string_of_field, lst));
    }
    
    switch t {
    case Int:	    return "int";
    case Bool:	    return "bool";
    case Void:	    return "void";
    case String:    return "string";
    case Float:	    return "float";
    case Exn:	    return "exn";
    case AbstractType:	return "abstract-type";
    case Tuple[lst]:
	let args = List::map(string_of_type, lst) in
	return "*[" + String::concat(", ", args) + "]";
    case Func[ft]:
	let args = List::map(string_of_type, List::map(fst, ft.parms)) in
	return "*(" + string_of_type(ft.ret_type) + 
	       " (" + String::concat(", ", args) + "))";
    case TyVar[n]:
	return "'" + n;
	
    case XTyVar[x]:
	let hd = "'" + x.name + " /* = " in
	switch x.t {
	case Some[tt]:
	    return hd + string_of_type(tt) + " */";
	case None:
	    return hd + "? */";
	}

    case Struct[lst]:
	return "struct {" + record_content(lst) + "}";

    case OptStruct[lst]:
	return "opt_struct {" + record_content(lst) + "}";

    case Union[lst]:
	return "union {" + record_content(lst) + "}";
    
    case NamedType[nt]:
	let args = List::map(string_of_type, nt.parms) in
	let qn = qualified_name(nt.tydef) in
	switch args {
	case List::Nil:
	    return qn;
	case List::Cons[_, _]:	
	    return "<" + String::concat(", ", args) + ">" + qn;
	}
    }
}

bool has_weak_tyvars(ty t)
{
    bool has_xt = false;

    void f(ty tt)
    {
	switch tt {
	case XTyVar[xt]:
	    switch xt.t {
	    case None:
		has_xt = true;
	    case Some[x]:
		if (!has_xt && has_weak_tyvars(x))
		    has_xt = true;
	    }
	case _:
	    skip;
	}
    }

    type_iter(f, t);

    return has_xt;
}

bool is_base_type(ty t)
{
    switch t {
    case NamedType[nt]:
	let ts = get_binding(nt.tydef) in
	switch ts.stype {
	case Struct[_]:	    return true;
	case OptStruct[_]:  return true;
	case Union[_]:	    return true;
	case AbstractType:  return true;
	case NamedType[_]:  return false;
	case _:		    return false;
	}
	
    case _:
	return true;
    }
}

*['a, 'b] make_pair('a a, 'b b)
{
    return [a, b];
}

*[<xtyvar>List::t, ty] subst_xtyvars(ty t, <id>List::t ns)
{
    <xtyvar>List::t xtyvars = 
	List::map(fun xtyvar (id n) (({name = n, t = None})), ns);
	
    ty f(ty x)
    {
	switch x {
	case TyVar[a]:
	    try {
		id n = a;
		bool fnd(xtyvar xt) { return xt.name == n; }
		return XTyVar[List::find(fnd, xtyvars)];
	    } with {
	    case Not_found:
		return x;
	    }
	case _:
	    return x;
	}
    }

    return [xtyvars, type_map(f, t)];
}

ty subst_vars(ty t, <*[id, ty]>List::t assoc)
{
    let [xts, new_t] = subst_xtyvars(t, List::map(fst, assoc)) in {
	void f(xtyvar x, ty s) { x.t = Some[s]; }
	List::iter2(f, xts, List::map(snd, assoc));
	return new_t;
    }
}

ty type_instance_one(ty t)
{
    switch t {
    case NamedType[nt]:
	let ts = get_binding(nt.tydef) in
	let assoc = List::map2(make_pair, ts.a.tysym_tyvars, nt.parms) in
	return subst_vars(ts.stype, assoc);
    
    case XTyVar[xt]:
	switch xt.t {
	case Some[tt]:
	    return tt;
	case None:
	    return t;
	}
	
    case _:
	return t;
    }
}

bool is_type_instance(ty t)
{
    switch t {
    case NamedType[_]:
	return false;
    
    case XTyVar[xt]:
	switch xt.t {
	case Some[_]:
	    return false;
	case None:
	    return true;
	}
	
    case _:
	return true;
    }
}

ty type_instance(ty t)
{
    while (!is_type_instance(t))
	t = type_instance_one(t);

    return t;
}

ty base_type(ty t)
{
    while (!is_base_type(t))
	t = type_instance_one(t);
    
    return t;
}

void unify(ty t1, ty t2)
{
    void bail_out(string m)
    {
	raise Type_error["cannot unify " + string_of_type(t1) + " with " + 
			 string_of_type(t2) + m];
    }

//dbg("UN: " + string_of_type(t1) + " <-> " + string_of_type(t2) + "\n");
    
    switch [t1, t2] {
    case [Int, Int]:	    return;
    case [Void, Void]:	    return;
    case [Bool, Bool]:	    return;
    case [Float, Float]:    return;
    case [String, String]:  return;
    case [Exn, Exn]:	    return;

    case [Tuple[l1], Tuple[l2]]:
	if (List::length(l1) != List::length(l2))
	    bail_out(" because they have different number of arguments");
	List::iter2(unify, l1, l2);
    
    case [Func[f1], Func[f2]]:
	unify(f1.ret_type, f2.ret_type);
	if (List::length(f1.parms) != List::length(f2.parms))
	    bail_out(" because they have different number of arguments");
	List::iter2(unify, List::map(fst, f1.parms), 
		   List::map(fst, f2.parms));
    
    case [TyVar[v1], TyVar[v2]]:
	if (v1 != v2)
	    raise Type_error["cannot unify type variables '" + v1 + 
			     " and '" + v2];
    
    case [XTyVar[x1], XTyVar[x2]]:
	if (Hacks::ptr_of(t1) != Hacks::ptr_of(t2)) {
	    switch [x1.t, x2.t] {
	    case [Some[tt1], Some[tt2]]:
		unify(tt1, tt2);
		
	    case [Some[tt1], None]:
		x2.t = Some[tt1];
		
	    case [None, Some[tt2]]:
		x1.t = Some[tt2];
	    
	    case [None, None]:
		let t = XTyVar[({t = None, name = "weak"})] in {
		    x1.t = Some[t];
		    x2.t = Some[t];
		}
	    }
	}
	
    case [XTyVar[x], s2]:
	switch x.t {
	case Some[s1]:
	    unify(s1, s2);
	case None:
	    x.t = Some[s2];
	}
	
    case [s1, XTyVar[x]]:
	switch x.t {
	case Some[s2]:
	    unify(s1, s2);
	case None:
	    x.t = Some[s1];
	}
    
    // opt_struct {} is type of null
    case [OptStruct[List::Nil], OptStruct[_]]: skip; 
    
    case [OptStruct[_], OptStruct[List::Nil]]: skip;
    
    case [OptStruct[List::Nil], _]:
	if (is_type_instance(t2))
	    bail_out("");
	unify(t1, type_instance(t2));
	
    case [_, OptStruct[List::Nil]]:
	if (is_type_instance(t1))
	    bail_out("");
	unify(type_instance(t1), t2);

    // named-type/base-type stuff
    case [NamedType[nt1], NamedType[nt2]]:
	bool ss = same_symbols(get_binding(nt1.tydef), get_binding(nt2.tydef));

	if (ss) {
	    List::iter2(unify, nt1.parms, nt2.parms);
	} else {
	    if (!is_base_type(t1) || !is_base_type(t2)) {
		unify(base_type(t1), base_type(t2));
	    } else {
		bail_out("");
	    }
	}
    
    case [NamedType[_], _]:
	if (!is_base_type(t1))
	    unify(base_type(t1), t2);
	else
	    bail_out("");

    case [_, NamedType[_]]:
	if (!is_base_type(t2))
	    unify(t1, base_type(t2));
	else
	    bail_out("");
    
    case [AbstractType, AbstractType]:
	// slightly better error message then below
	raise Type_error["cannot unify different abstract types"];
	
    case [_, _]:
	bail_out("");
    }
}

ty type_of_s(expr e)
{
    return type_instance(type_of(e));
}

bool is_void_func_type(ty t)
{
    let ft = get_ft_record(type_instance(t)) in
    switch ft.ret_type {
    case Void:
	return true;
    case _:
	return false;
    }
}

void tycheck_vardef(vardef v)
{
    switch type_instance((get_binding(v.var_symbol)).stype) {
    case Void:
	error("cannot have variable of void type");
    case _:
	skip;
    }
}

struct ty_ctx {
    ty fnc_ret_type;
}

void tycheck_lambda(ty_ctx c, lambda_record ll)
{
    let ft = get_ft_record(ll.lambda_type) in
    let cc = ({fnc_ret_type = ft.ret_type}) in
    tycheck_stmt(cc, ll.lambda_body);
}

bool is_lvalue(expr e)
{
    switch e.a.raw {
    case Ref[rr]:
	return has_property(get_binding(rr.ref_symbol), IsLvalue);

    case DotExpr[de]:
	return true;
	
    case TupleExpr[l]:
	return !List::exists(fun bool (expr ex) (!is_lvalue(ex)), l);
	
    case _:
	return false;
    }
}

void expect(string msg, ty t, expr e)
{
    try {
	unify(t, type_of(e));
    } with {
    case Type_error[m]:
	error("`" + msg + "' expects type " + string_of_type(t) + ", not " + 
	      string_of_type(type_of(e)));
	if (flags.verbose_errors)
	    error("(" + m + ")");
	raise Choke;
    }
}

int unxt_cnt;

ty unxt(ty tt)
{
    int lev = 0;
    
    ty f(ty t)
    {
	switch t {
	case XTyVar[xt]:
	    switch xt.t {
	    case Some[x]:
		if (lev++ > 100)
		    ice("Type::unxt(): cyclic type");
		t = f(x);
		lev--;
		return t;
	    case None:
		if (String::sub(xt.name, 0, 1) != "_")
		    xt.name = "_" + itoa(++unxt_cnt);
		return t;
	    }
	case _:
	    return t;
	}
    }

    return type_map(f, tt);
}

ty type_of_xtyvar(xtyvar xt)
{
    return unxt(XTyVar[xt]);
}

ty freshen_lambda(ty t)
{
    switch type_instance(t) {
    case Func[ft]:
	let [_, tt] = subst_xtyvars(t, ft.tyvars) in
	return tt;
    case _:
	return t;
    }
}

ty tycheck_call(ty_ctx c, ty fnc, <ty>List::t raw_args)
{
//dbg("tycheck-call: " + string_of_type(fnc) + "\n");
    let args = List::map(freshen_lambda, raw_args) in
    let xt = XTyVar[({name = "rt", t = None})] in
    let parms = List::map(fun *[ty, <id>option] (ty x) ([x, None]), args) in
    let formal_ty = Func[({ret_type = xt, parms = parms, tyvars = List::Nil})] in 
    let ti = type_instance(fnc) in
    switch ti {
    case Func[ft]:
	let [_, fresh_fnc] = subst_xtyvars(fnc, ft.tyvars) in
	try {
	    unify(fresh_fnc, formal_ty);
	} with {
	case Type_error[m]:
	    error("type error in call: " + m);
	    raise Choke;
	}

	return unxt(xt);

    case XTyVar[_]:
	try {
	    unify(ti, formal_ty);
	} with {
	case Type_error[_]:
	    ice("Type::tycheck_call(): failed to unify weak");
	}

	return unxt(xt);
	
    case t:
	error("expression of non-functional type " + string_of_type(t) + 
	      " called");
	raise Choke;
    }
}

*[<xtyvar>List::t, ty] fresh_tydef_content(ty_symbol ts)
{
    return subst_xtyvars(type_instance(ts.stype), ts.a.tysym_tyvars);
}

ty substituted_tydef(<xtyvar>List::t xtyvars, ty_symbol ts)
{
    let parms = List::map(type_of_xtyvar, xtyvars) in
    return NamedType[({parms = parms, tydef = Symbol[ts]})];
}

ty_symbol union_of(var_symbol label)
{
    <ty_symbol>option parent_union_o = None;
    void check_parent_union(var_symbol_property p)
    {
	switch p {
	case ParentUnion[u]:
	    parent_union_o = Some[u];
	case _:
	    skip;
	}
    }
    
    List::iter(check_parent_union, label.a.properties);
    
    switch parent_union_o {
    case Some[ts]:
	return ts;
    case None:
	return ice("Type::union_of(): not union");
    }
}

ty type_of_union_field_in(ty t, <var_symbol>bindable name)
{
    id sn = simple_name(name);
    
    bool is_our_fld(<var_symbol>field f) 
    {
	return simple_name(f.fld_symbol) == sn;
    }
    
    <var_symbol>field fld;

    switch type_instance(t) {
    case Union[content]:
	try {
	    fld = List::find(is_our_fld, content);
	} with {
	case Not_found:
	    error("`" + sn + "' is not part of " + string_of_type(t));
	    raise Choke;
	}

	return fld.fld_type;
	
    case _:
	return ice("Type::type_of_union_field_in(): not union");
    }
}

ty tycheck_union_expr(ty_ctx c, namedexpr nx)
{
    int tuple_len(ty t)
    {
	switch type_instance(t) {
	case Tuple[lst]:
	    return List::length(lst);
	case Void:
	    return 0;
	case _:
	    return 1;
	}
    }
    
    void check_with(ty t)
    {
	let ex_type = type_of(nx.nx_expr) in
	let ex_len = tuple_len(ex_type) in
	let t_len = tuple_len(t) in
	try {
	    switch [type_instance(t), nx.nx_expr.a.raw] {
	    case [Void, VoidConst]:
		skip;	// ok
	    case [Void, _]:
		// we don't really want ,,void f(); List::Nil[f()]''
		raise Type_error["void type constructor is given a value"];
	    case _:
		skip;
	    }
	
	    if (ex_len != t_len) {
		let m = qualified_name(nx.nx_symbol) + " takes " + itoa(t_len) + 
		    " argument(s), not " + itoa(ex_len) in
		raise Type_error[m];
	    }
	    unify(ex_type, t);
	} with {
	case Type_error[m]:
	    error("type error in union expression: " + m);
	    raise Choke;
	}
    }

    // first of all we typecheck the expression
    tycheck_expr(c, nx.nx_expr);
    
    // next we need to find out parent union
    var_symbol v = get_binding(nx.nx_symbol);

    if (has_property(v, IsException)) {
	check_with(v.stype);
	return Exn;
    }
    
    ty_symbol parent_union = union_of(v);

    // and finally unify type of our expression with type of
    // our union field
    let tyvars = parent_union.a.tysym_tyvars in
    let [xtvars, t] = subst_xtyvars(v.stype, tyvars) in {
	check_with(t);
	return substituted_tydef(xtvars, parent_union);
    }
}

ty tycheck_struct_expr(ty_ctx c, <sini_expr>List::t inis)
{
    ty_symbol ts;
    
    switch inis {
    case List::Cons[hd, _]:
	ts = (get_binding(hd.sx_symbol)).a.sf_tydef;
    case List::Nil:
	return ice("Type::tycheck_struct_expr(): empty sx");
    }
    
    void check_one(<sf_symbol>field fld)
    {
	id sn = simple_name(fld.fld_symbol);
	sini_expr se;

	try {
	    se = List::find(fun bool (sini_expr s) 
				(simple_name(s.sx_symbol) == sn), 
			    inis);
	} with {
	case Not_found:
	    error("struct field `" + qualified_name(fld.fld_symbol) +
		  "' not initalized");
	    return;
	}
	
	tycheck_expr(c, se.sx_expr);
	try {
//dbg("x: " + string_of_type(type_of(se.sx_expr)) + " f: " + string_of_type(fld.fld_type) + "\n");
	    unify(type_of(se.sx_expr), fld.fld_type);
	} with {
	case Type_error[m]:
	    error("type error: " + m + " in initalization of struct field `" +
		  qualified_name(se.sx_symbol) + "'");
	}
    }
    
    let [xtyvars, t] = fresh_tydef_content(ts) in {
	switch t {
	case Struct[x]:
	    List::iter(check_one, x);
	    
	case OptStruct[x]:
	    List::iter(check_one, x);
	    
	case _:
	    return ice("Type::tycheck_struct_expr(): non-struct");
	}

	return substituted_tydef(xtyvars, ts);
    }
}

ty tycheck_dot_expr(ty_ctx c, sini_expr de)
{
    tycheck_expr(c, de.sx_expr);

    <<sf_symbol>field>List::t content;

    switch type_of_s(de.sx_expr) {
    case Struct[x]:
	content = x;
    case OptStruct[x]:
	content = x;
    case _:
	error("left-hand side of `.' is of non-struct type " +
	      string_of_type(type_of(de.sx_expr)));
	raise Choke;
    }

    id sn = simple_name(de.sx_symbol);
    <sf_symbol>field fld;

    bool is_our_fld(<sf_symbol>field f) 
    {
	return simple_name(f.fld_symbol) == sn;
    }
    
    try {
	fld = List::find(is_our_fld, content);
    } with {
    case Not_found:
	error("`" + sn + "' is not part of " + 
	      string_of_type(type_of(de.sx_expr)));
	raise Choke;
    }

    // Bind it. Binder couldn't do that.
    de.sx_symbol = Symbol[get_binding(fld.fld_symbol)];

    return fld.fld_type;
}

bool is_constant_expr(expr e)
{
    switch e.a.raw {
    case SimpleBinOp[_, e1, e2]:
	return is_constant_expr(e1) && is_constant_expr(e2);
	
    case SimpleUnOp[_, e1]:
	return is_constant_expr(e1);
    
    case Ref[rr]:
	if (has_property(get_binding(rr.ref_symbol), IsUnionField))
	    return true;
	return false;

    case Call[_, _]:
	return false;

    case TupleExpr[lst]:
	return !List::exists(fun bool (expr x) (!is_constant_expr(x)), lst);

    case UnionExpr[u]:
	return is_constant_expr(u.nx_expr);
    
    case Lambda[_]:
    case StructExpr[_]:
    case DotExpr[_]:
	return false;
	
    case BoolConst[_]:
    case IntConst[_]:
    case FloatConst[_]:
    case StringConst[_]:
    case VoidConst:
    case NullConst:
	return true;
    }
}

// TODO: fix to recognize XTyVar
void tycheck_pattern(ty_ctx c, ty expr_type, pattern pat)
{
    void pat_unify(ty t, pattern p)
    {
	push_location(p.loc);
	
	try {
	    switch p.a {
	    case VoidPattern:
		skip;
	    case ExprPattern[e]:
		tycheck_expr(c, e);
		try {
		    unify(type_of(e), t);
		} with {
		case Type_error[m]:
		    error("type error with constant pattern: " + m);
		    raise Choke;
		}
		
		if (!is_constant_expr(e)) {
		    error("pattern expression is not constant");
		    raise Choke;
		}
	    
	    case VarPattern[v]:
		try {
		    unify(v.var_type, t);
		} with {
		case Type_error[m]:
		    error("type error with pattern variable: " + m);
		}
	    
	    case TuplePattern[lst]:
		if (List::length(lst) < 2)
		    ice("Bind::tycheck_pattern(): tuple shorter then 2");

		switch type_instance(t) {
		case Tuple[tlst]:
		    try {
			List::iter2(pat_unify, tlst, lst);
		    } with {
		    case Invalid_argument["List::iter2"]:
			error("number of arguments in tuple pattern and tuple " + 
			      "type doesn't match");
			raise Choke;
		    }
		case _:
		    error("tuple pattern used to match non-tuple type " + 
			  string_of_type(t));
		}
	    
	    case UnionPattern[n]:
		let sym = get_binding(n.np_symbol) in
		if (has_property(sym, IsException)) {
		    switch type_instance(t) {
		    case Exn:
			pat_unify(sym.stype, n.np_pattern);
			return;
		    case _:
			error("this pattern matches exceptions, not " +
			      string_of_type(t));
			raise Choke;
		    }
		    
		} 
		
		switch type_instance(t) {
		case Union[List::Cons[f, _]]:
		    let eu = union_of(get_binding(f.fld_symbol)) in
		    let pu = union_of(get_binding(n.np_symbol)) in
		    if (!same_symbols(eu, pu)) {
			error("pattern matching union " + symbol_name(pu) + 
			      " is used to match union " + symbol_name(eu));
			raise Choke;
		    }
		    
		case _:
		    let pu = union_of(get_binding(n.np_symbol)) in {
			error("pattern matching union " + symbol_name(pu) +
			      " is used to match values of non-union type " + 
			      string_of_type(t));
			raise Choke;
		    }
		}

		let et = type_of_union_field_in(t, n.np_symbol) in
		pat_unify(et, n.np_pattern);
	    }
	} finally {
	    pop_location();
	}
    }

    pat_unify(expr_type, pat);
}

void tycheck_case_clause(ty_ctx c, ty expr_type, case_clause cc)
{
    void tycheck_one_case(*[pattern, expr] oc)
    {
	let [pat, cond] = oc in {
	    tycheck_pattern(c, expr_type, pat);
	    try {
		tycheck_expr(c, cond);
		unify(Bool, type_of(cond));
	    } with {
	    case Type_error[m]:
		errorl(cond.loc, "case guard needs to have bool type, but " + m);
	    }
	}
    }

    List::iter(tycheck_one_case, cc.cases);
    tycheck_stmt(c, cc.body);
}

ty tycheck_binop(ty_ctx c, op2 op, expr e1, expr e2)
{
    ty t; // this is set by helper routines below

    void same_arith_types(string msg, expr ex1, expr ex2)
    {
	let t1 = type_of_s(ex1) in
	let t2 = type_of_s(ex2) in
	switch [t1, t2] {
	case [Int, Int]:
	case [Float, Float]:	
	    skip;
	    
	case [Int, Float]:
	    error("`" + msg + 
		  "' won't do int -> float conversion on left argument");
	    if (flags.verbose_errors)
		error("type conversions in Gont have to be explicit");
	    t = Float;
	    
	case [Float, Int]:
	    error("`" + msg + 
		  "' won't do int -> float conversion on right argument");
	    if (flags.verbose_errors)
		error("type conversions in Gont have to be explicit");
	    t = Float;
	    
	case [Int, XTyVar[_]]:
	case [Float, XTyVar[_]]:
	case [XTyVar[_], Float]:
	case [XTyVar[_], Int]:
	    unify(t1, t2);
	    t = type_instance(t1);
	    
	// better error message
	case [XTyVar[_], XTyVar[_]]:
	    error("sorry, cannot infer types of arguments of `" + msg + "'");
	    raise Choke;
	    
	case _:
	    error("`" + msg + "' works on ints and floats, not on " +
		  string_of_type(t1) + " and " + string_of_type(t2));
	    raise Choke;
	}
    }

    void tycheck_cmp(string sop, expr ex1, expr ex2)
    {
	t = Bool;

	let t1 = type_of_s(ex1) in
	let t2 = type_of_s(ex2) in
	switch [t1, t2] {
	case [Int, Int]:
	case [Float, Float]:
	case [String, String]:
	    skip;
	    
	case [Int, XTyVar[_]]:
	case [Float, XTyVar[_]]:
	case [String, XTyVar[_]]:
	case [XTyVar[_], Int]:
	case [XTyVar[_], String]:
	case [XTyVar[_], Float]:
	    unify(t1, t2);
	    
	// better error message
	case [XTyVar[_], XTyVar[_]]:
	    error("sorry, cannot infer types of arguments of `" + sop + "'");
	    
	case _:
	    error("`" + sop + "' cannot work on " + string_of_type(t1) + 
		  " and " + string_of_type(t2));
	}
    }
    
    void tycheck_cmp_eq(string sop, expr ex1, expr ex2)
    {
	t = Bool;
	
	let t1 = type_of_s(ex1) in
	let t2 = type_of_s(ex2) in
	switch [t1, t2] {
	// null checks
	case [OptStruct[List::Nil], OptStruct[_]]:
	case [OptStruct[_], OptStruct[List::Nil]]:
	    skip;
	    
	// bools are ok
	case [Bool, Bool]:
	case [XTyVar[_], Bool]:
	case [Bool, XTyVar[_]]:
	    unify(t1, t2);
	    
	// and other _cmp stuff:
	case _:
	    tycheck_cmp(sop, ex1, ex2);
	}
    }
    
    void tycheck_plus(expr ex1, expr ex2)
    {
	// cheat a bit, + takes the same argument types as < does
	tycheck_cmp("+", ex1, ex2);
	t = type_of_s(ex1);
    }
    
    tycheck_expr(c, e1);
    tycheck_expr(c, e2);

    t = type_of(e1);

    switch op {
    case Plus:	tycheck_plus(e1, e2);
    case Minus:	same_arith_types("-", e1, e2);
    case Mult:	same_arith_types("*", e1, e2);
    case Div:	same_arith_types("/", e1, e2);
    case Mod:
	expect("%", Int, e1);
	expect("%", Int, e2);
    case Lt:	tycheck_cmp("<", e1, e2);
    case Gt:	tycheck_cmp(">", e1, e2);
    case Ge:	tycheck_cmp(">=", e1, e2);
    case Le:	tycheck_cmp("<=", e1, e2);
    case Eq:	tycheck_cmp_eq("==", e1, e2);
    case Ne:	tycheck_cmp_eq("!=", e1, e2);
    case AndIf:
	expect("&&", Bool, e1);
	expect("&&", Bool, e2);
    case OrIf:
	expect("||", Bool, e1);
	expect("||", Bool, e2);
    case Assign:
	try {
	    unify(type_of(e1), type_of(e2));
	} with {
	case Type_error[m]:
	    error("incompatible types in assigment: " + m);
	}

	if (!is_lvalue(e1))
	    error("assigment to non-lvalue");
    }

    return t;
}

void tycheck_raw_expr(ty_ctx c, nl_expr ex)
{
    ty t;

    ty self(expr e)
    {
	tycheck_expr(c, e);
	return type_of(e);
    }
    
    switch ex.raw {
    case SimpleBinOp[op, e1, e2]:
	t = tycheck_binop(c, op, e1, e2);
	
    case SimpleUnOp[op, e]:
	tycheck_expr(c, e);

	t = type_of(e);
	
	switch op {
	case Negate:
	    switch (type_of_s(e)) {
	    case Int:
		skip;
	    case Float:
		skip;
	    case _:
		error("negate operator expects integer or " +
		      "float argument, not " + string_of_type(type_of(e)));
	    }
	case PreDec:	expect("pre --", Int, e);
	case PreInc:	expect("pre ++", Int, e);
	case PostDec:	expect("post --", Int, e);
	case PostInc:	expect("post ++", Int, e);
	}
    
    case Ref[rr]:
	t = (get_binding(rr.ref_symbol)).stype;

    case Call[e, el]:
	tycheck_expr(c, e);
	let arg_types = List::map(self, el) in
	t = tycheck_call(c, type_of(e), arg_types);

    case TupleExpr[el]:
	t = Tuple[List::map(self, el)];

    case UnionExpr[u]:
	tycheck_expr(c, u.nx_expr);
	t = tycheck_union_expr(c, u);
    
    case Lambda[l]:
	tycheck_lambda(c, l);
	t = l.lambda_type;

    case StructExpr[s]:
	t = tycheck_struct_expr(c, s);

    case DotExpr[de]:
	t = tycheck_dot_expr(c, de);
    
    case BoolConst[_]:	    t = Bool;
    case IntConst[_]:	    t = Int;
    case FloatConst[_]:	    t = Float;
    case StringConst[_]:    t = String;
    case VoidConst:	    t = Void;
    case NullConst:	    t = OptStruct[List::Nil];
    }

    ex.expr_type = Some[t];
}

void tycheck_expr(ty_ctx c, expr e)
{
    locate(partial(tycheck_raw_expr, c), e);
}


void tycheck_stmt_nl(ty_ctx c, nl_stmt s)
{
    switch s {
    case Skip:
	skip;
    
    case Expr[e]:
	tycheck_expr(c, e);
    
    case Return[e]:
	tycheck_expr(c, e);
	unify(type_of(e), c.fnc_ret_type);

    case VarDef[v]:
	tycheck_vardef(v);
	
    case Block[lst]:
	List::iter(partial(tycheck_stmt, c), lst);

    case Pair[s1, s2]:
	tycheck_stmt(c, s1);
	tycheck_stmt(c, s2);
    
    case Switch[e, clauses]:
	tycheck_expr(c, e);
	List::iter(partial2(tycheck_case_clause, c, type_of(e)), clauses);
	
    case Let[assigns, st]:
	void check_one(*[pattern, expr] a)
	{
	    let [p, e] = a in {
		tycheck_expr(c, e);
		tycheck_pattern(c, type_of(e), p);
	    }
	}

	List::iter(check_one, assigns);
	tycheck_stmt(c, st);
    
    case If[e, s1, s2]:
	tycheck_expr(c, e);
	unify(Bool, type_of(e));
	tycheck_stmt(c, s1);
	tycheck_stmt(c, s2);
    
    case Loop[l]:
	tycheck_stmt(c, l.ll_body);
	tycheck_stmt(c, l.ll_inc);
    
    case Break[_]:
	skip;
	
    case BreakIfNot[e]:
	tycheck_expr(c, e);
	unify(Bool, type_of(e));
    
    case Open[_]:
	skip;
    
    case Try[tr]:
	tycheck_stmt(c, tr.try_body);
	List::iter(partial2(tycheck_case_clause, c, Exn), tr.try_with);
	tycheck_stmt(c, tr.try_finally);
    
    case Raise[e]:
	tycheck_expr(c, e);
	unify(Exn, type_of(e));
    }
}

void tycheck_stmt(ty_ctx ctx, stmt s)
{
    void check(nl_stmt ss)
    {
	try {
	    tycheck_stmt_nl(ctx, ss);
	} with {
	case Type_error[m]:
	    error(m);
	}
    }
    
    locate(check, s);
}

void check_var_with_interface(var_symbol impl)
{
    var_symbol iface;
    
    try {
	iface = Module::lookup_var(impl.module, impl.name);
    } with {
    case Not_found:
	return;
    }
    
    try {
	unify(iface.stype, impl.stype);
    } with {
    case Type_error[s]:
	error("implementation of " + symbol_name(impl) + 
	      " is incompatible with type declared in interface: " + s);
	errorl(iface.sloc, "type in interface is: " + 
	       string_of_type(iface.stype));
    }

    // set this for Codegen, so it will know to put (public) in front of it
    set_property(impl, IsExported);
    // and this for ourselves so we can check whatever all variables
    // declared were defined
    set_property(iface, IsExported);
}

void check_module_completness()
{
    if (Bind::doing_iface)
	return;
    
    Module::module m = Module::open_module(Bind::current_module);

    <var_symbol>List::t missing = List::Nil;

    void checker(string dummy, var_symbol s)
    {
	if (!has_property(s, IsExported) &&
	    !has_property(s, IsUnionField) &&
	    !has_property(s, IsExternC))
	    missing = List::Cons[s, missing];
    }

    Hashtbl::iter(m.mod_vars, checker);

    if (List::length(missing) != 0) {
	error("symbols declared in interface but not implemented:");
	void print(var_symbol s)
	{
	    errorl(s.sloc, "- `" + s.name + "' of type " +
		   string_of_type(s.stype));
	}
	List::iter(print, missing);
    }
}

void tycheck_nl_decl(nl_decl nld)
{
    ty_ctx mk_ty_ctx(ty ret)
    {
	return ({fnc_ret_type = ret});
    }

    switch nld {
    case FuncDecl[fn]:
	let s = get_binding(fn.fn_symbol) in {
	    if (has_weak_tyvars(s.stype)) {
		error("global symbols cannot have weak type variables");
		raise Choke;
	    }
	    check_var_with_interface(s);
	}
	// Use bogus context, typcheck_lambda() is expected
	// to make its own.
	tycheck_lambda(mk_ty_ctx(AbstractType), fn.fn_lambda);
    
    case SectionDecl[_, l]:
	// likewise
	tycheck_lambda(mk_ty_ctx(AbstractType), l);
    
    case TopLevelOpen[_]:
	skip;
	
    case GlobalDecl[v]:
	if (!Bind::doing_iface)
	    check_var_with_interface(get_binding(v.var_symbol));
	tycheck_vardef(v);
	
    case ExternC[v]:
	tycheck_vardef(v);
	
    case TyDef[_]:
	// redefinition and compatibility with interface was already
	// checked by binder
	skip;
    
    case FuncProto[_]:
    case XCFuncProto[_, _]:
    case ExceptionDecl[_]:
    case DummyDecl:
	skip;
    }
}

void tycheck_decl(decl d)
{
    locate(tycheck_nl_decl, d);
}

bool is_global_function(var_symbol s)
{
    switch type_instance(s.stype) {
    case Func[_]:
    return s.a.fnc_level <= 0 && !has_property(s, IsGlobalVar);
    case _:
    return false;
    }
}

// vim:sts=4:sw=4
